#!/usr/bin/python3

import os
import sys
import cv2
import numpy as np
from tflite_runtime.interpreter import Interpreter
import argparse
import time
from math import ceil, sqrt, exp, pi, floor, sin, cos, atan2, gcd
import mediapipe_utils as mpu

DIR_NAME = os.path.dirname(__file__)
LANDMARK_MODEL = 'hand.tflite'
DETECTION_MODEL = 'palm.tflite'

DEMO_WINDOW_NAME = 'Hand Estimation Demo'

# layout settings for armadillo-demo-experience
CAMERA_WIDTH = 1100
CAMERA_HEIGHT = 600
OVERLAY_CAMERA_XPOS = 0
OVERLAY_CAMERA_YPOS = 0

class HandRegion:
    def __init__(self, pd_score=None, pd_box=None, pd_kps=None):
        self.pd_score = pd_score # Palm detection score
        self.pd_box = pd_box # Palm detection box [x, y, w, h] normalized
        self.pd_kps = pd_kps # Palm detection keypoints

    def get_rotated_world_landmarks(self):
        world_landmarks_rotated = self.world_landmarks.copy()
        sin_rot = sin(self.rotation)
        cos_rot = cos(self.rotation)
        rot_m = np.array([[cos_rot, sin_rot], [-sin_rot, cos_rot]])
        world_landmarks_rotated[:,:2] = np.dot(world_landmarks_rotated[:,:2], rot_m)
        return world_landmarks_rotated

    def print(self):
        attrs = vars(self)
        print('\n'.join("%s: %s" % item for item in attrs.items()))

class HandLandmarker:
    def __init__(self, model_path):
        self.interpreter = Interpreter(model_path)
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        self.input_h = self.input_details[0]['shape'][1]
        self.input_w = self.input_details[0]['shape'][2]

    def infer(self, input_image):
        h, w, _ = input_image.shape
        out_img = input_image.copy()

        img = self._preprocess(input_image)
        self.interpreter.set_tensor(self.input_details[0]['index'], img)
        start = time.monotonic()
        self.interpreter.invoke()
        elapsed_time = time.monotonic() - start
        fps_text = 'hand_landmark FPS(inference only): {:.2f}'.format(1 / elapsed_time)

        output = self.interpreter.get_tensor(self.output_details[2]['index'])

        cv_keypoints, norm_landmarks = self._postprocess(output, w, h)

        return (cv_keypoints, norm_landmarks, fps_text)

    def _preprocess(self, input_image):
        image = cv2.resize(input_image, (self.input_h, self.input_w))
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = image / 255
        image = np.expand_dims(image, 0)
        image = image.astype(self.input_details[0]['dtype'])
        return image

    def _postprocess(self, output, frame_width, frame_height):
        h_scale = frame_height / self.input_h
        w_scale = frame_width / self.input_w
        keypoint_coords = np.reshape(np.squeeze(output), (21, 3))
        cv_keypoints = []
        for p in keypoint_coords:
            cv_keypoints.append(cv2.KeyPoint(p[0] * w_scale, p[1] * h_scale, 10.))
        return (cv_keypoints, (keypoint_coords / self.input_h))

    def draw_hand(self, img, keypoints):
        base = (int(keypoints[0].pt[0]), int(keypoints[0].pt[1]))
        fingers = np.asarray(keypoints[1:]).reshape(5, 4)
        for finger in fingers:
            cv2.line(
                img,
                base,
                (int(finger[0].pt[0]), int(finger[0].pt[1])),
                (0, 0, 255),
                thickness=2
            )
            for i in range(len(finger) - 1):
                cv2.line(
                    img,
                    (int(finger[i].pt[0]), int(finger[i].pt[1])),
                    (int(finger[i+1].pt[0]), int(finger[i+1].pt[1])),
                    (0, 0, 255),
                    thickness=2
                )
        img = cv2.drawKeypoints(
            img,
            keypoints,
            outImage=np.array([]),
            color=(255, 0, 0),
            flags=cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS
        )
        return img

class PalmDetector:
    def __init__(self, model_path, pd_nms_thresh=0.3, max_hands=2):
        self.interpreter = Interpreter(model_path)
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        self.input_h = self.input_details[0]['shape'][1]
        self.input_w = self.input_details[0]['shape'][2]

        self.i_scale, self.i_zeropoint = self.input_details[0]['quantization']
        self.o_scale0, self.o_zeropoint0 = self.output_details[0]['quantization']
        self.o_scale1, self.o_zeropoint1 = self.output_details[1]['quantization']

        self.scale = self.input_h
        self.pd_input_length = self.input_h
        out_len = self.output_details[0]['shape'][1]
        self.anchors = mpu.generate_handtracker_anchors(self.pd_input_length, self.pd_input_length)[:out_len]
        self.pd_nms_thresh = pd_nms_thresh
        self.max_hands = max_hands

    def infer(self, input_image):
        h, w, _ = input_image.shape
        img = self._preprocess(input_image)
        self.interpreter.set_tensor(self.input_details[0]['index'], img)
        start = time.monotonic()
        self.interpreter.invoke()
        elapsed_time = time.monotonic() - start
        fps_text = '   palm_detect FPS(inference only): {:.2f}'.format(1 / elapsed_time)
        return (self._postprocess(w, h), fps_text)

    def _preprocess(self, input_image):
        image = cv2.resize(input_image, (self.input_h, self.input_w))
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = image / 255
        if self.i_scale != 0:
            image = (image / self.i_scale) + self.i_zeropoint
        image = np.expand_dims(image, 0)
        image = image.astype(self.input_details[0]['dtype'])
        return image

    def _postprocess(self, frame_width, frame_height):
        scores = self.interpreter.get_tensor(self.output_details[0]['index'])
        scores = np.squeeze(scores)
        bboxes = self.interpreter.get_tensor(self.output_details[1]['index'])
        bboxes = np.squeeze(bboxes)
        if self.o_scale0 != 0 and self.o_scale1 != 0:
            scores = (scores + self.o_zeropoint0) + self.o_scale0
            bboxes = (bboxes + self.o_zeropoint1) + self.o_scale1

        hands = mpu.decode_bboxes(0.93, scores, bboxes, self.anchors, scale=self.pd_input_length, best_only=False)
        if len(hands) > 0:
            mpu.detections_to_rect(hands)
            mpu.rect_transformation(hands, frame_width, frame_height)
            if len(hands) >= 2:
                hands = mpu.non_max_suppression(hands, self.pd_nms_thresh)[:self.max_hands]
            return hands
        else:
            return None

class Inferencer:
    def __init__(self, landmark_model, detection_model, camera):
        self.hand_landmarker = HandLandmarker(os.path.join(DIR_NAME, landmark_model))
        self.palm_detector = PalmDetector(os.path.join(DIR_NAME, detection_model))
        self.camera = camera

    def _crop(self, frame, rect_points):
        h, w, _ = frame.shape
        left = w
        right = 0
        top = h
        bottom = 0
        for p in rect_points:
            left = max(0, min(p[0], left))
            right = min(w, max(p[0], right))
            top = max(0, min(p[1], top))
            bottom = min(h, max(p[1], bottom))
        return (frame[top:bottom, left:right], (left, top))

    def _overlay_imgs(self, bg, fg, xpos, ypos):
        frame = bg
        height, width = fg.shape[:2]
        frame[ypos:height + ypos, xpos:width + xpos] = fg
        return frame

    def _scale_world(self, keypoints, trans, center, crp_center):
        rotated_points = []
        dist_x = center[0] - crp_center[0]
        dist_y = center[1] - crp_center[1]
        for k in keypoints:
            p = np.array([[k.pt[0] + dist_x], [k.pt[1] + dist_y], [1]])
            rotated_p = trans.dot(p)
            rotated_points.append(cv2.KeyPoint(int(rotated_p[0]), int(rotated_p[1]), k.size))
        return rotated_points

    def infer(
            self,
            target_window,
            cw,
            ch,
            ocx=OVERLAY_CAMERA_XPOS,
            ocy=OVERLAY_CAMERA_YPOS,
            back_ground_image=None
    ):
        if back_ground_image is None:
            back_ground_image = np.zeros((ch, cw, 3), np.uint8)
            ocx = 0
            ocy = 0

        while True:
            ret, frame = self.camera.read()
            if not ret:
                sys.exit(
                    'ERROR: Unable to read from webcam. Please verify your webcam settings.'
                )
            f = frame.copy()
            h, w, _ = frame.shape

            hands, detector_fps = self.palm_detector.infer(frame)
            if not hands is None:
                world_key_points = []
                for hand in hands:
                    hand_center = (int(hand.rect_x_center_a), int(hand.rect_y_center_a))
                    f = cv2.polylines(
                        f,
                        [np.array(hand.rect_points)],
                        True,
                        color=(0, 255, 0),
                        thickness=3,
                        lineType=cv2.LINE_4
                    )
                    cropped, coord = self._crop(frame, hand.rect_points)
                    trans = cv2.getRotationMatrix2D(
                        hand_center,
                        np.rad2deg(hand.rotation),
                        1
                    )
                    trans_inv = cv2.getRotationMatrix2D(
                        hand_center,
                        np.rad2deg(-hand.rotation),
                        1
                    )
                    ft = cv2.warpAffine(f, trans, (w, h))
                    cropped = cv2.getRectSubPix(
                        ft,
                        (int(hand.rect_w_a), int(hand.rect_h_a)),
                        hand_center
                    )
                    crp_center = (cropped.shape[1] // 2, cropped.shape[0] // 2)
                    kp, nl, landmarker_fps = self.hand_landmarker.infer(cropped)
                    wkp = self._scale_world(
                        kp,
                        trans_inv,
                        hand_center,
                        crp_center
                    )
                    f = self.hand_landmarker.draw_hand(f, wkp)

                    hand.norm_landmarks = nl
                    mpu.recognize_gesture(hand)
                    if not hand.gesture is None:
                        g_text_size = 1.5
                        (tw_g, th_g), _ = cv2.getTextSize(hand.gesture, cv2.FONT_HERSHEY_SIMPLEX, g_text_size, 1)
                        text_pos = (int(hand_center[0] - (tw_g // 2)), int(hand_center[1] - (hand.rect_h_a // 2)))
                        cv2.rectangle(
                            f,
                            (text_pos[0] - 7, text_pos[1] + 7),
                            (text_pos[0] + tw_g + 7, text_pos[1] - (th_g + 7)),
                            (255, 255, 255),
                            thickness=-1
                        )
                        putText(f, hand.gesture, text_pos, g_text_size)

                (tw_d, th_d), _ = cv2.getTextSize(detector_fps, cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)
                cv2.rectangle(f, (10, 2), (tw_d + 10, th_d + 7), (255, 255, 255), thickness=-1)
                putText(f, detector_fps, (10, 15))
                (tw_l, th_l), _ = cv2.getTextSize(landmarker_fps, cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)
                cv2.rectangle(f, (10, th_d + 5), (tw_l + 10, th_d + 5 + th_l + 7), (255, 255, 255), thickness=-1)
                putText(f, landmarker_fps, (10, th_d + 20))
            f = self._overlay_imgs(
                back_ground_image,
                cv2.resize(f, dsize=(cw, ch)),
                ocx,
                ocy
            )
            cv2.imshow(DEMO_WINDOW_NAME, f)
            cv2.waitKey(1)

def putText(image, text, org, scale=0.5):
    cv2.putText(
        image,
        text=text,
        org=org,
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=scale,
        color=(0, 0, 255),
        thickness=1,
        lineType=cv2.LINE_AA
    )

def resize_back_ground(image_path, width, height):
    image = cv2.imread(image_path)
    image = cv2.resize(image, dsize=(width, height))
    w_scale = width / DEF_WINDOW_WIDTH
    h_scale = height / DEF_WINDOW_HEIGHT
    return (image, w_scale, h_scale)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--landmark_model',
        type=str,
        default=LANDMARK_MODEL
    )
    parser.add_argument(
        '--detection_model',
        type=str,
        default=DETECTION_MODEL
    )
    parser.add_argument(
        '--camera_id',
        type=int,
        default=2
    )
    parser.add_argument(
        '--width',
        type=int,
        default=CAMERA_WIDTH
    )
    parser.add_argument(
        '--height',
        type=int,
        default=CAMERA_HEIGHT
    )
    args = parser.parse_args()

    cap = cv2.VideoCapture(args.camera_id, cv2.CAP_V4L2)
    cap.set(cv2.CAP_PROP_FRAME_WIDTH, args.width)
    cap.set(cv2.CAP_PROP_FRAME_HEIGHT, args.height)

    inferencer = Inferencer(args.landmark_model, args.detection_model, cap)
    try:
        inferencer.infer(
            target_window=DEMO_WINDOW_NAME,
            cw=args.width,
            ch=args.height
        )
    except KeyboardInterrupt:
        cap.release()
        cv2.destroyAllWindows()

if __name__ == '__main__':
    main()
