import os
import cv2
import argparse
import sys
import threading
import queue
import numpy as np
from yolov5_tflite_inference import yolov5_tflite

WINDOW_NAME = 'screw_detection'

MODEL_NAME = 'screw_detection.tflite'
DIR_NAME = os.path.dirname(__file__)

# layout settings for armadillo-demo-experience
CAMERA_WIDTH = 1100
CAMERA_HEIGHT = 600

class FrameHandler:
    def __init__(
            self,
            len_thres,
            imgsrc,
            visualize_score=False,
            visualize_length=False
    ):
        self.picture_flag = False
        if type(imgsrc) is str:
            self.current_image = cv2.imread(imgsrc)
            self.camera = None
        else:
            self.camera = imgsrc
            self.current_image = None
        self.len_thres = len_thres
        self.visualize_score = visualize_score
        self.visualize_length = visualize_length
        self.infer_result = None

    def set_infer_result(self, data):
        self.infer_result = data

    def _calc_distance(self, box):
        return np.linalg.norm(box[2:] - box[:2])

    def run(self, cw, ch):
        while True:
            if not self.camera is None:
                ret, self.current_image = self.camera.read()
                if not ret:
                    print('video end.')
                    sys.exit(0)
            frame = cv2.resize(self.current_image.copy(), (cw, ch))

            if self.infer_result is None:
                continue
            boxes, scores, class_names, pred_time = self.infer_result

            short_count = 0
            for box, score in zip(boxes, scores):
                box = box.astype(np.int32)
                length = self._calc_distance(box)
                if length <= self.len_thres:
                    color = (0, 0, 255)
                    short_count += 1
                else:
                    color = (255, 0, 0)
                cv2.rectangle(frame, box[:2], box[2:], color, 2)
                box_text = ''
                if self.visualize_score:
                    box_text += '{:.2f} '.format(score)
                if self.visualize_length:
                    box_text += '{}'.format(int(length))
                if box_text != '':
                    (tw, th), _ = cv2.getTextSize(box_text, cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)
                    cv2.rectangle(
                        frame,
                        box[:2],
                        (box[0] + (tw + 3), box[1] - (th + 2)),
                        color,
                        thickness=-1
                    )
                    cv2.putText(
                        frame,
                        text=box_text,
                        org=box[:2],
                        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                        fontScale=0.5,
                        color=(255, 255, 255),
                        thickness=1,
                        lineType=cv2.LINE_AA
                    )

            font_scale = 2
            font_face = cv2.FONT_HERSHEY_DUPLEX
            pred_text = 'FPS(inference only): {:.2f}'.format(1 / pred_time)
            (tw_d, th_d), _ = cv2.getTextSize(pred_text, font_face, font_scale, 1)
            cv2.rectangle(frame, (10, 2), (tw_d + 10, th_d + 7), (255, 255, 255), thickness=-1)
            cv2.putText(
                frame,
                text=pred_text,
                org=(10, 50),
                fontFace=font_face,
                fontScale=font_scale,
                color=(0, 0, 255),
                thickness=1,
                lineType=cv2.LINE_AA
            )

            count_text = 'screws: {}'.format(len(boxes))
            short_text = 'short screws: {}'.format(short_count)
            (tw_d, th_d), _ = cv2.getTextSize(count_text, font_face, font_scale, 1)
            text_org = [cw - 600, ch - 60]
            cv2.rectangle(
                frame,
                (text_org[0] + 193, text_org[1] - th_d),
                (text_org[0] + tw_d + 5 + 193, text_org[1] + th_d - 8),
                (255, 255, 255),
                thickness=-1
            )
            cv2.putText(
                frame,
                text=count_text,
                org=(text_org[0] + 193, text_org[1]),
                fontFace=font_face,
                fontScale=font_scale,
                color=(0, 0, 255),
                thickness=1,
                lineType=cv2.LINE_AA
            )
            (tw_l, th_l), _ = cv2.getTextSize(short_text, font_face, font_scale, 1)
            text_org[1] += th_d + 5
            cv2.rectangle(
                frame,
                (text_org[0], text_org[1] - th_l),
                (text_org[0] + tw_l + 5, text_org[1] + th_l - 8),
                (255, 255, 255),
                thickness=-1
            )
            cv2.putText(
                frame,
                text=short_text,
                org=text_org,
                fontFace=font_face,
                fontScale=font_scale,
                color=(0, 0, 255),
                thickness=1,
                lineType=cv2.LINE_AA
            )

            cv2.imshow(WINDOW_NAME, frame)
            cv2.waitKey(1)


    def get_frame(self):
        return self.current_image


class ScrewDetector(yolov5_tflite): 
    def __init__(self, model_path, conf_thres, iou_thres, imgsrc):
        super().__init__(
            model_path,
            conf_thres=conf_thres,
            iou_thres=iou_thres
        )
        self.imgsrc = imgsrc
        self.q = queue.Queue()

    def run(self, cw, ch):
        while True:
            frame = self.imgsrc.get_frame()
            if frame is None:
                continue
            frame = cv2.resize(frame, (cw, ch))
            self.q.put(self.detect(frame))

    def get_infer_result(self):
        return self.q.get()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--model',
        type=str,
        default=MODEL_NAME
    )
    parser.add_argument(
        '--input_img',
        type=str
    )
    parser.add_argument(
        '--input_video',
        type=str
    )
    parser.add_argument(
        '--camera_id',
        type=int,
        default=2
    )
    parser.add_argument(
        '--width',
        type=int,
        default=CAMERA_WIDTH
    )
    parser.add_argument(
        '--height',
        type=int,
        default=CAMERA_HEIGHT
    )
    parser.add_argument(
        '--conf_thres',
        type=float,
        default=0.5
    )
    parser.add_argument(
        '--iou_thres',
        type=float,
        default=0.45
    )
    parser.add_argument(
        '--len_thres',
        type=int,
        default=130
    )
    parser.add_argument(
        '--visualize_score',
        action='store_true'
    )
    parser.add_argument(
        '--visualize_length',
        action='store_true'
    )
    args = parser.parse_args()

    model_path = os.path.join(DIR_NAME, args.model)

    cv2.namedWindow(WINDOW_NAME, cv2.WINDOW_NORMAL)

    if not args.input_img is None:
        src = args.input_img
    elif not args.input_video is None:
        src = cv2.VideoCapture(args.input_video)
    else:
        src = cv2.VideoCapture(args.camera_id, cv2.CAP_V4L2)
        src.set(cv2.CAP_PROP_FRAME_WIDTH, CAMERA_WIDTH)
        src.set(cv2.CAP_PROP_FRAME_HEIGHT, CAMERA_HEIGHT)

    frame_handler = FrameHandler(
        args.len_thres,
        imgsrc=src,
        visualize_score=args.visualize_score,
        visualize_length=args.visualize_length
    )

    t_fhandler = threading.Thread(
        target=frame_handler.run,
        args=(args.width, args.height)
    )
    t_fhandler.start()

    inferencer = ScrewDetector(
        model_path=model_path,
        conf_thres=args.conf_thres,
        iou_thres=args.iou_thres,
        imgsrc=frame_handler
    )

    t_inferencer = threading.Thread(
        target=inferencer.run,
        args=(args.width, args.height)
    )
    t_inferencer.start()

    while True:
        frame_handler.set_infer_result(inferencer.get_infer_result())


if __name__ == '__main__':
    main()
