from tflite_runtime.interpreter import Interpreter
import os
import sys
import cv2
import numpy as np
import argparse
import time

DIR_NAME = os.path.dirname(__file__)

MODEL_NAME = 'super_resolution.tflite'

DEMO_WINDOW_NAME = 'Super Resolution Demo'

# layout settings for armadillo-demo-experience
CAMERA_WIDTH = 1100
CAMERA_HEIGHT = 600
OVERLAY_CAMERA_XPOS = 0
OVERLAY_CAMERA_YPOS = 0

DEF_WINDOW_WIDTH = 1920
DEF_WINDOW_HEIGHT = 1080
BG_FRAME_WIDTH = 320
BG_FRAME_HEIGHT = 180

DEF_FONT_SIZE = 22
DEF_FONT_MARGIN = 2

class Inferencer:
    def __init__(self, model_path, camera):
        self.interpreter = Interpreter(model_path)
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        _, self.input_height, self.input_width, _ = self.input_details[0]['shape']
        _, self.output_height, self.output_width, _ = self.output_details[0]['shape']
        self.input_dtype = self.input_details[0]['dtype']
        self.camera = camera

    def _preprocess_input(self, image):
        image_for_prediction = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image_for_prediction = image_for_prediction.astype(self.input_dtype)
        return np.expand_dims(image_for_prediction, 0)

    def _postprocess_output(self, output):
        output = np.squeeze(output)
        output = np.clip(output, 0, 255)
        output = np.round(output)
        output = output.astype(np.uint8)
        return cv2.cvtColor(output, cv2.COLOR_RGB2BGR)

    def _crop(self, image):
        img_height, img_width, _ = image.shape
        self.crop_top = int(img_height / 2 - self.input_height / 2)
        self.crop_bottom = self.crop_top + self.input_height
        self.crop_left = int(img_width / 2 - self.input_width / 2)
        self.crop_right = self.crop_left + self.input_width
        return image[self.crop_top:self.crop_bottom, self.crop_left:self.crop_right].copy()

    def _overlay_imgs(self, bg, fg, xpos, ypos):
        frame = bg
        height, width = fg.shape[:2]
        frame[ypos:height + ypos, xpos:width + xpos] = fg
        return frame

    def infer(
            self,
            target_window,
            cw,
            ch,
            ocx=OVERLAY_CAMERA_XPOS,
            ocy=OVERLAY_CAMERA_YPOS,
            back_ground_image=None
    ):
        if back_ground_image is None:
            back_ground_image = np.zeros((ch, cw, 3), np.uint8)
            ocx = 0
            ocy = 0

        while True:
            success, image = self.camera.read()
            if not success:
                sys.exit(
                    'ERROR: Unable to read from webcam. Please verify your webcam settings.'
                )

            cropped_img = self._crop(image)
            input_img = self._preprocess_input(cropped_img)

            start = time.monotonic()
            self.interpreter.set_tensor(self.input_details[0]['index'], input_img)
            self.interpreter.invoke()
            elapsed_time = time.monotonic() - start

            res = self.interpreter.get_tensor(self.output_details[0]['index'])
            output = self._postprocess_output(res)

            # overlay center rectangle
            cv2.rectangle(image, (self.crop_left, self.crop_top), (self.crop_right, self.crop_bottom), (0, 0, 255), thickness=2)
            crop_size_text = '50 x 50'
            (cstw, csth), _ = cv2.getTextSize(crop_size_text, cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)
            cstx = self.crop_left + int(self.input_width / 2 - cstw / 2)
            cv2.rectangle(image, (cstx, self.crop_bottom + 2), (cstx + cstw, self.crop_bottom + csth + 4), (255, 255, 255), thickness=-1)
            putText(image, crop_size_text, (cstx, self.crop_bottom + csth + DEF_FONT_MARGIN))

            # overlay captured image
            frame = self._overlay_imgs(
                back_ground_image,
                cv2.resize(image, dsize=(cw, ch)),
                ocx,
                ocy
            )

            # overlay resized image (INTER NEAREST)
            ri_xpos = cw + ocx - int(ch / 2)
            ri_ypos = 0
            frame = self._overlay_imgs(
                    frame,
                    cv2.resize(cropped_img, dsize=(int(ch / 2), int(ch / 2)), interpolation=cv2.INTER_NEAREST),
                    ri_xpos,
                    ri_ypos
            )
            fscale = cw / DEF_WINDOW_WIDTH
            org_image_text = 'UPSCALE ORIGINAL'
            (tw, th), _ = cv2.getTextSize(org_image_text, cv2.FONT_HERSHEY_SIMPLEX, fscale, 1)
            cv2.rectangle(frame, (ri_xpos, DEF_FONT_MARGIN), (ri_xpos + tw, th + DEF_FONT_MARGIN * 2), (255, 255, 255), thickness=-1)
            putText(frame, org_image_text, (ri_xpos, th + DEF_FONT_MARGIN), fscale)

            # overlay super resolution image
            sr_xpos = ri_xpos
            sr_ypos = int(ch / 2)
            frame = self._overlay_imgs(
                    frame,
                    cv2.resize(output, dsize=(int(ch / 2), int(ch / 2))),
                    sr_xpos,
                    sr_ypos
            )
            sr_image_text = 'SUPER RESOLUTION'
            fps_text = 'FPS(inference only): {:.2f}'.format(1 / elapsed_time)
            (_, srth), _ = cv2.getTextSize(sr_image_text, cv2.FONT_HERSHEY_SIMPLEX, fscale, 1)
            (fpstw, fpsth), _ = cv2.getTextSize(fps_text, cv2.FONT_HERSHEY_SIMPLEX, fscale, 1)
            cv2.rectangle(frame, (ri_xpos, sr_ypos), (ri_xpos + fpstw, sr_ypos + srth + fpsth + DEF_FONT_MARGIN * 4), (255, 255, 255), thickness=-1)
            putText(frame, sr_image_text, (sr_xpos, sr_ypos + srth + DEF_FONT_MARGIN), fscale)
            putText(frame, fps_text, (sr_xpos, sr_ypos + srth + fpsth + 2 * DEF_FONT_MARGIN), fscale)

            cv2.imshow(target_window, frame)
            cv2.waitKey(1)

def putText(image, text, org, scale=0.5):
    cv2.putText(
        image,
        text=text,
        org=org,
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=scale,
        color=(0, 0, 255),
        thickness=1,
        lineType=cv2.LINE_AA
    )

def resize_back_ground(image_path, width, height):
    image = cv2.imread(image_path)
    image = cv2.resize(image, dsize=(width, height))
    w_scale = width / DEF_WINDOW_WIDTH
    h_scale = height / DEF_WINDOW_HEIGHT
    return (image, w_scale, h_scale)

def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '--model',
        help='File path of .tflite file.',
        required=False,
        default=MODEL_NAME
    )
    parser.add_argument(
        '--camera_id',
        type=int,
        default=2
    )
    parser.add_argument(
        '--width',
        type=int,
        default=CAMERA_WIDTH
    )
    parser.add_argument(
        '--height',
        type=int,
        default=CAMERA_HEIGHT
    )
    parser.add_argument(
        '--file',
        type=str,
        default=None,
        help="Optionally use a video file instead of a live camera"
    )

    args = parser.parse_args()

    if args.file is not None:
        cap = cv2.VideoCapture(args.file)
    else:
        cap = cv2.VideoCapture(args.camera_id + cv2.CAP_V4L2)
    cap.set(cv2.CAP_PROP_FRAME_WIDTH, args.width)
    cap.set(cv2.CAP_PROP_FRAME_HEIGHT, args.height)

    model = os.path.join(DIR_NAME, args.model)
    inferencer = Inferencer(model, cap)

    try:
        inferencer.infer(
            target_window=DEMO_WINDOW_NAME,
            cw=args.width,
            ch=args.height
        )
    except KeyboardInterrupt:
        cap.release()
        cv2.destroyAllWindows()

if __name__ == '__main__':
    main()
